@@ -129,6 +129,9 @@ BaremetalGroup = [
     cfg.ListOpt('enabled_rescue_interfaces',
                 default=['no-rescue'],
                 help="List of Ironic enabled rescue interfaces."),
+    cfg.ListOpt('enabled_boot_interfaces',
+                default=['fake', 'pxe'],
+                help="List of Ironic enabled boot interfaces."),
     cfg.StrOpt('default_rescue_interface',
                help="Ironic default rescue interface."),
     cfg.IntOpt('adjusted_root_disk_size_gb',
