@@ -20,6 +20,24 @@ class BaremetalClient(base.BaremetalClient):
     version = '1'
     uri_prefix = 'v1'
 
+    @staticmethod
+    def _get_headers(api_version):
+        """Return headers for a request.
+
+        Currently supports a header specifying the API version to use.
+
+        :param api_version: Ironic API version to use.
+        :return: a 2-tuple of (extra_headers, headers), where 'extra_headers'
+            is whether to use headers, and 'headers' is a list of headers to
+            use in the request.
+        """
+        extra_headers = False
+        headers = None
+        if api_version is not None:
+            extra_headers = True
+            headers = {'x-openstack-ironic-api-version': api_version}
+        return extra_headers, headers
+
     @base.handle_errors
     def list_nodes(self, **kwargs):
         """List all existing nodes."""
@@ -70,6 +88,11 @@ class BaremetalClient(base.BaremetalClient):
         """List all existing states."""
         return self._list_request('/nodes/%s/states' % uuid)
 
+    @base.handle_errors
+    def list_node_bios_settings(self, uuid):
+        """List node bios settings."""
+        return self._list_request('/nodes/%s/bios' % uuid)
+
     @base.handle_errors
     def list_ports_detail(self, **kwargs):
         """Details list all existing ports."""
@@ -81,28 +104,33 @@ class BaremetalClient(base.BaremetalClient):
         return self._list_request('drivers')
 
     @base.handle_errors
-    def show_node(self, uuid):
+    def show_node(self, uuid, api_version=None):
         """Gets a specific node.
 
         :param uuid: Unique identifier of the node in UUID format.
+        :param api_version: Ironic API version to use.
         :return: Serialized node as a dictionary.
 
         """
-        return self._show_request('nodes', uuid)
+        extra_headers, headers = self._get_headers(api_version)
+        return self._show_request('nodes', uuid, headers=headers,
+                                  extra_headers=extra_headers)
 
     @base.handle_errors
-    def show_node_by_instance_uuid(self, instance_uuid):
+    def show_node_by_instance_uuid(self, instance_uuid, api_version=None):
         """Gets a node associated with given instance uuid.
 
         :param instance_uuid: Unique identifier of the instance in UUID format.
+        :param api_version: Ironic API version to use.
         :return: Serialized node as a dictionary.
 
         """
         uri = '/nodes/detail?instance_uuid=%s' % instance_uuid
-
+        extra_headers, headers = self._get_headers(api_version)
         return self._show_request('nodes',
                                   uuid=None,
-                                  uri=uri)
+                                  uri=uri, headers=headers,
+                                  extra_headers=extra_headers)
 
     @base.handle_errors
     def show_chassis(self, uuid):
@@ -254,9 +282,8 @@ class BaremetalClient(base.BaremetalClient):
             uuid: UUID of the port group. Optional.
         :return: A tuple with the server response and the created port group.
         """
-        portgroup = {'extra': kwargs.get('extra', {'foo': 'bar'})}
-
-        portgroup['node_uuid'] = node_uuid
+        portgroup = {'extra': kwargs.get(
+            'extra', {'foo': 'bar', 'open': 'stack'}), 'node_uuid': node_uuid}
 
         if kwargs.get('address'):
             portgroup['address'] = kwargs['address']
@@ -266,6 +293,17 @@ class BaremetalClient(base.BaremetalClient):
 
         return self._create_request('portgroups', portgroup)
 
+    @base.handle_errors
+    def update_portgroup(self, uuid, patch):
+        """Update the specified port group.
+
+        :param uuid: The unique identifier of the port group.
+        :param patch: List of dicts representing json patches.
+        :return: A tuple with the server response and the updated port group.
+        """
+
+        return self._patch_request('portgroups', uuid, patch)
+
     @base.handle_errors
     def create_volume_connector(self, node_uuid, **kwargs):
         """Create a volume connector with the specified parameters.
@@ -382,18 +420,25 @@ class BaremetalClient(base.BaremetalClient):
         :return: A tuple with the server response and the updated node.
 
         """
+        if 'reset_interfaces' in kwargs:
+            params = {'reset_interfaces': str(kwargs.pop('reset_interfaces'))}
+        else:
+            params = {}
+
         node_attributes = ('properties/cpu_arch',
                            'properties/cpus',
                            'properties/local_gb',
                            'properties/memory_mb',
                            'driver',
+                           'bios_interface',
                            'deploy_interface',
+                           'rescue_interface',
                            'instance_uuid',
                            'resource_class')
         if not patch:
             patch = self._make_patch(node_attributes, **kwargs)
 
-        return self._patch_request('nodes', uuid, patch)
+        return self._patch_request('nodes', uuid, patch, params=params)
 
     @base.handle_errors
     def update_chassis(self, uuid, **kwargs):
@@ -460,7 +505,7 @@ class BaremetalClient(base.BaremetalClient):
 
     @base.handle_errors
     def set_node_provision_state(self, node_uuid, state, configdrive=None,
-                                 clean_steps=None):
+                                 clean_steps=None, rescue_password=None):
         """Set provision state of the specified node.
 
         :param node_uuid: The unique identifier of the node.
@@ -469,6 +514,7 @@ class BaremetalClient(base.BaremetalClient):
         :param configdrive: A gzipped, base64-encoded
             configuration drive string.
         :param clean_steps: A list with clean steps to execute.
+        :param rescue_password: user password used to rescue.
         """
         data = {'target': state}
         # NOTE (vsaienk0): Add both here if specified, do not check anything.
@@ -477,6 +523,8 @@ class BaremetalClient(base.BaremetalClient):
             data['configdrive'] = configdrive
         if clean_steps is not None:
             data['clean_steps'] = clean_steps
+        if rescue_password is not None:
+            data['rescue_password'] = rescue_password
         return self._put_request('nodes/%s/states/provision' % node_uuid,
                                  data)
 
@@ -640,3 +688,58 @@ class BaremetalClient(base.BaremetalClient):
         resp, body = self.get(uri)
         self.expected_success(200, resp.status)
         return resp, self.deserialize(body)
+
+    @base.handle_errors
+    def list_node_traits(self, node_uuid):
+        """List all traits associated with the node.
+
+        :param node_uuid: The unique identifier of the node.
+        """
+        return self._list_request('/nodes/%s/traits' % node_uuid)
+
+    @base.handle_errors
+    def set_node_traits(self, node_uuid, traits):
+        """Set all traits of the specified node.
+
+        :param node_uuid: The unique identifier of the node.
+        :param traits: A list of traits to set.
+        """
+        request = {'traits': traits}
+        resp, body = self._put_request('nodes/%s/traits' %
+                                       node_uuid, request)
+        self.expected_success(http_client.NO_CONTENT, resp.status)
+        return resp, body
+
+    @base.handle_errors
+    def add_node_trait(self, node_uuid, trait):
+        """Add a trait to the specified node.
+
+        :param node_uuid: The unique identifier of the node.
+        :param trait: A trait to add.
+        """
+        resp, body = self._put_request('nodes/%s/traits/%s' %
+                                       (node_uuid, trait), {})
+        self.expected_success(http_client.NO_CONTENT, resp.status)
+        return resp, body
+
+    @base.handle_errors
+    def remove_node_traits(self, node_uuid):
+        """Remove all traits from the specified node.
+
+        :param node_uuid: Unique identifier of the node in UUID format.
+        """
+        resp, body = self._delete_request('nodes/%s/traits' % node_uuid, {})
+        self.expected_success(http_client.NO_CONTENT, resp.status)
+        return resp, body
+
+    @base.handle_errors
+    def remove_node_trait(self, node_uuid, trait):
+        """Remove a trait from the specified node.
+
+        :param node_uuid: Unique identifier of the node in UUID format.
+        :param trait: A trait to remove.
+        """
+        resp, body = self._delete_request('nodes/%s/traits/%s' %
+                                          (node_uuid, trait), {})
+        self.expected_success(http_client.NO_CONTENT, resp.status)
+        return resp, body
